@@ -19,6 +19,7 @@ import matplotlib.pyplot as plt
 import matplotlib.transforms as mtransforms
 import numpy as np
 import numpy.ma as ma
+import numpy.typing as npt
 import pandas as pd
 import shapely
 import spatialdata as sd
@@ -54,12 +55,21 @@ from skimage.color import label2rgb
 from skimage.morphology import erosion, square
 from skimage.segmentation import find_boundaries
 from skimage.util import map_array
-from spatialdata import SpatialData, get_element_annotators, get_extent, get_values, rasterize
-from spatialdata._core.query.relational_query import _locate_value, _ValueOrigin
+from spatialdata import (
+    SpatialData,
+    get_element_annotators,
+    get_extent,
+    get_values,
+    rasterize,
+)
+from spatialdata._core.query.relational_query import _locate_value
 from spatialdata._types import ArrayLike
-from spatialdata.models import Image2DModel, Labels2DModel, PointsModel, SpatialElement, get_model
+from spatialdata.models import Image2DModel, Labels2DModel, SpatialElement
+
+# from spatialdata.transformations.transformations import Scale
+from spatialdata.transformations import Affine, Identity, MapAxis, Scale, Translation
+from spatialdata.transformations import Sequence as SDSequence
 from spatialdata.transformations.operations import get_transformation
-from spatialdata.transformations.transformations import Scale
 from xarray import DataArray, DataTree
 
 from spatialdata_plot._logging import logger
@@ -178,7 +188,12 @@ def _prepare_params_plot(
     dpi = rcParams["figure.dpi"] if dpi is None else dpi
     if num_panels > 1 and ax is None:
         fig, grid = _panel_grid(
-            num_panels=num_panels, hspace=hspace, wspace=wspace, ncols=ncols, dpi=dpi, figsize=figsize
+            num_panels=num_panels,
+            hspace=hspace,
+            wspace=wspace,
+            ncols=ncols,
+            dpi=dpi,
+            figsize=figsize,
         )
         axs: None | Sequence[Axes] = [plt.subplot(grid[c]) for c in range(num_panels)]
     elif num_panels > 1:
@@ -301,7 +316,6 @@ def _get_centroid_of_pathpatch(pathpatch: mpatches.PathPatch) -> tuple[float, fl
 
 
 def _scale_pathpatch_around_centroid(pathpatch: mpatches.PathPatch, scale_factor: float) -> None:
-
     centroid = _get_centroid_of_pathpatch(pathpatch)
     vertices = pathpatch.get_path().vertices
     scaled_vertices = np.array([centroid + (vertex - centroid) * scale_factor for vertex in vertices])
@@ -381,7 +395,11 @@ def _get_collection_shape(
     shapes_df = shapes_df.reset_index(drop=True)
 
     def _assign_fill_and_outline_to_row(
-        fill_c: list[Any], outline_c: list[Any], row: dict[str, Any], idx: int, is_multiple_shapes: bool
+        fill_c: list[Any],
+        outline_c: list[Any],
+        row: dict[str, Any],
+        idx: int,
+        is_multiple_shapes: bool,
     ) -> None:
         try:
             if is_multiple_shapes and len(fill_c) == 1:
@@ -397,7 +415,10 @@ def _get_collection_shape(
         coords = np.array(row["geometry"].exterior.coords)
         centroid = np.mean(coords, axis=0)
         scaled_coords = (centroid + (coords - centroid) * s).tolist()
-        return {**row.to_dict(), "geometry": mpatches.Polygon(scaled_coords, closed=True)}
+        return {
+            **row.to_dict(),
+            "geometry": mpatches.Polygon(scaled_coords, closed=True),
+        }
 
     def _process_multipolygon(row: pd.Series, s: float) -> list[dict[str, Any]]:
         mp = _make_patch_from_multipolygon(row["geometry"])
@@ -505,6 +526,8 @@ def _prepare_cmap_norm(
 
     cmap = copy(cmap)
 
+    assert isinstance(cmap, Colormap), f"Invalid type of `cmap`: {type(cmap)}, expected `Colormap`."
+
     if norm is None:
         norm = Normalize(vmin=None, vmax=None, clip=False)
 
@@ -673,7 +696,7 @@ def _get_colors_for_categorical_obs(
             palette = default_102
         else:
             palette = ["grey" for _ in range(len_cat)]
-            logger.info("input has more than 103 categories. Uniform " "'grey' color will be used for all categories.")
+            logger.info("input has more than 103 categories. Uniform 'grey' color will be used for all categories.")
     else:
         # raise error when user didn't provide the right number of colors in palette
         if isinstance(palette, list) and len(palette) != len(categories):
@@ -699,17 +722,6 @@ def _get_colors_for_categorical_obs(
     return palette[:len_cat]  # type: ignore[return-value]
 
 
-# TODO consider move to relational query in spatialdata
-def get_values_point_table(sdata: SpatialData, origin: _ValueOrigin, table_name: str) -> pd.Series:
-    """Get a particular column stored in _ValueOrigin from the table in the spatialdata object."""
-    table = sdata[table_name]
-    if origin.origin == "obs":
-        return table.obs[origin.value_key]
-    if origin.origin == "var":
-        return table[:, table.var_names.isin([origin.value_key])].X.copy()
-    raise ValueError(f"Color column `{origin.value_key}` not found in table {table_name}")
-
-
 def _set_color_source_vec(
     sdata: sd.SpatialData,
     element: SpatialElement | None,
@@ -719,14 +731,22 @@ def _set_color_source_vec(
     groups: list[str] | str | None = None,
     palette: list[str] | str | None = None,
     cmap_params: CmapParams | None = None,
+    alpha: float = 1.0,
     table_name: str | None = None,
+    table_layer: str | None = None,
+    render_type: Literal["points"] | None = None,
 ) -> tuple[ArrayLike | pd.Series | None, ArrayLike, bool]:
     if value_to_plot is None and element is not None:
         color = np.full(len(element), na_color)
         return color, color, False
 
     # Figure out where to get the color from
-    origins = _locate_value(value_key=value_to_plot, sdata=sdata, element_name=element_name, table_name=table_name)
+    origins = _locate_value(
+        value_key=value_to_plot,
+        sdata=sdata,
+        element_name=element_name,
+        table_name=table_name,
+    )
 
     if len(origins) > 1:
         raise ValueError(
@@ -734,13 +754,13 @@ def _set_color_source_vec(
         )
 
     if len(origins) == 1:
-        color_source_vector = _robust_get_value(
+        color_source_vector = get_values(
+            value_key=value_to_plot,
             sdata=sdata,
-            origin=origins[0],
-            value_to_plot=value_to_plot,
             element_name=element_name,
             table_name=table_name,
-        )
+            table_layer=table_layer,
+        )[value_to_plot]
 
         # numerical case, return early
         # TODO temporary split until refactor is complete
@@ -764,9 +784,12 @@ def _set_color_source_vec(
             adata=sdata.table,
             cluster_key=value_to_plot,
             color_source_vector=color_source_vector,
+            cmap_params=cmap_params,
+            alpha=alpha,
             groups=groups,
             palette=palette,
             na_color=na_color,
+            render_type=render_type,
         )
 
         color_source_vector = color_source_vector.set_categories(color_mapping.keys())
@@ -856,6 +879,7 @@ def _generate_base_categorial_color_mapping(
     cluster_key: str,
     color_source_vector: ArrayLike | pd.Series[CategoricalDtype],
     na_color: ColorLike,
+    cmap_params: CmapParams | None = None,
 ) -> Mapping[str, str]:
     if adata is not None and cluster_key in adata.uns and f"{cluster_key}_colors" in adata.uns:
         colors = adata.uns[f"{cluster_key}_colors"]
@@ -872,7 +896,7 @@ def _generate_base_categorial_color_mapping(
 
         return dict(zip(categories, colors, strict=True))
 
-    return _get_default_categorial_color_mapping(color_source_vector)
+    return _get_default_categorial_color_mapping(color_source_vector=color_source_vector, cmap_params=cmap_params)
 
 
 def _modify_categorical_color_mapping(
@@ -896,22 +920,36 @@ def _modify_categorical_color_mapping(
 
 def _get_default_categorial_color_mapping(
     color_source_vector: ArrayLike | pd.Series[CategoricalDtype],
+    cmap_params: CmapParams | None = None,
 ) -> Mapping[str, str]:
     len_cat = len(color_source_vector.categories.unique())
-    if len_cat <= 20:
-        palette = default_20
-    elif len_cat <= 28:
-        palette = default_28
-    elif len_cat <= len(default_102):  # 103 colors
-        palette = default_102
+    # Try to use provided colormap first
+    if cmap_params is not None and cmap_params.cmap is not None and not cmap_params.cmap_is_default:
+        # Generate evenly spaced indices for the colormap
+        color_idx = np.linspace(0, 1, len_cat)
+        if isinstance(cmap_params.cmap, ListedColormap):
+            palette = [to_hex(x) for x in cmap_params.cmap(color_idx)]
+        elif isinstance(cmap_params.cmap, LinearSegmentedColormap):
+            palette = [to_hex(cmap_params.cmap(x)) for x in color_idx]
+        else:
+            # Fall back to default palettes if cmap is not of expected type
+            palette = None
     else:
-        palette = ["grey" for _ in range(len_cat)]
-        logger.info("input has more than 103 categories. Uniform 'grey' color will be used for all categories.")
+        palette = None
 
-    return {
-        cat: to_hex(to_rgba(col)[:3])
-        for cat, col in zip(color_source_vector.categories, palette[:len_cat], strict=True)
-    }
+    # Fall back to default palettes if needed
+    if palette is None:
+        if len_cat <= 20:
+            palette = default_20
+        elif len_cat <= 28:
+            palette = default_28
+        elif len_cat <= len(default_102):  # 103 colors
+            palette = default_102
+        else:
+            palette = ["grey"] * len_cat
+            logger.info("input has more than 103 categories. Uniform 'grey' color will be used for all categories.")
+
+    return dict(zip(color_source_vector.categories, palette[:len_cat], strict=True))
 
 
 def _get_categorical_color_mapping(
@@ -919,8 +957,11 @@ def _get_categorical_color_mapping(
     na_color: ColorLike,
     cluster_key: str | None = None,
     color_source_vector: ArrayLike | pd.Series[CategoricalDtype] | None = None,
+    cmap_params: CmapParams | None = None,
+    alpha: float = 1,
     groups: list[str] | str | None = None,
     palette: list[str] | str | None = None,
+    render_type: Literal["points"] | None = None,
 ) -> Mapping[str, str]:
     if not isinstance(color_source_vector, Categorical):
         raise TypeError(f"Expected `categories` to be a `Categorical`, but got {type(color_source_vector).__name__}")
@@ -928,20 +969,41 @@ def _get_categorical_color_mapping(
     if isinstance(groups, str):
         groups = [groups]
 
+    if not palette and render_type == "points" and cmap_params is not None and not cmap_params.cmap_is_default:
+        palette = cmap_params.cmap
+
+        color_idx = color_idx = np.linspace(0, 1, len(color_source_vector.categories))
+        if isinstance(palette, ListedColormap):
+            palette = [to_hex(x) for x in palette(color_idx, alpha=alpha)]
+        elif isinstance(palette, LinearSegmentedColormap):
+            palette = [to_hex(palette(x, alpha=alpha)) for x in color_idx]  # type: ignore[attr-defined]
+        return dict(zip(color_source_vector.categories, palette, strict=True))
+
     if isinstance(palette, str):
         palette = [palette]
 
     if cluster_key is None:
         # user didn't specify a column to use for coloring
-        base_mapping = _get_default_categorial_color_mapping(color_source_vector)
+        base_mapping = _get_default_categorial_color_mapping(
+            color_source_vector=color_source_vector, cmap_params=cmap_params
+        )
     else:
-        base_mapping = _generate_base_categorial_color_mapping(adata, cluster_key, color_source_vector, na_color)
+        base_mapping = _generate_base_categorial_color_mapping(
+            adata=adata,
+            cluster_key=cluster_key,
+            color_source_vector=color_source_vector,
+            na_color=na_color,
+            cmap_params=cmap_params,
+        )
 
     return _modify_categorical_color_mapping(mapping=base_mapping, groups=groups, palette=palette)
 
 
 def _maybe_set_colors(
-    source: AnnData, target: AnnData, key: str, palette: str | ListedColormap | Cycler | Sequence[Any] | None = None
+    source: AnnData,
+    target: AnnData,
+    key: str,
+    palette: str | ListedColormap | Cycler | Sequence[Any] | None = None,
 ) -> None:
     color_key = f"{key}_colors"
     try:
@@ -1063,7 +1125,13 @@ def _get_list(
     raise ValueError(f"Can't make a list from variable: `{var}`")
 
 
-def save_fig(fig: Figure, path: str | Path, make_dir: bool = True, ext: str = "png", **kwargs: Any) -> None:
+def save_fig(
+    fig: Figure,
+    path: str | Path,
+    make_dir: bool = True,
+    ext: str = "png",
+    **kwargs: Any,
+) -> None:
     """
     Save a figure.
 
@@ -1355,7 +1423,12 @@ def _multiscale_to_spatial_image(
 
 
 def _get_elements_to_be_rendered(
-    render_cmds: list[tuple[str, ImageRenderParams | LabelsRenderParams | PointsRenderParams | ShapesRenderParams]],
+    render_cmds: list[
+        tuple[
+            str,
+            ImageRenderParams | LabelsRenderParams | PointsRenderParams | ShapesRenderParams,
+        ]
+    ],
     cs_contents: pd.DataFrame,
     cs: str,
 ) -> list[str]:
@@ -1429,7 +1502,15 @@ def _validate_show_parameters(
             f"the following strings: {readable_font_weights}.",
         )
 
-    font_sizes = ["xx-small", "x-small", "small", "medium", "large", "x-large", "xx-large"]
+    font_sizes = [
+        "xx-small",
+        "x-small",
+        "small",
+        "medium",
+        "large",
+        "x-large",
+        "xx-large",
+    ]
 
     if legend_fontsize is not None and (
         not isinstance(legend_fontsize, int | float | str)
@@ -1522,7 +1603,11 @@ def _type_check_params(param_dict: dict[str, Any], element_type: str) -> dict[st
     if (contour_px := param_dict.get("contour_px")) and not isinstance(contour_px, int):
         raise TypeError("Parameter 'contour_px' must be an integer.")
 
-    if (color := param_dict.get("color")) and element_type in {"shapes", "points", "labels"}:
+    if (color := param_dict.get("color")) and element_type in {
+        "shapes",
+        "points",
+        "labels",
+    }:
         if not isinstance(color, str):
             raise TypeError("Parameter 'color' must be a string.")
         if element_type in {"shapes", "points"}:
@@ -1606,13 +1691,13 @@ def _type_check_params(param_dict: dict[str, Any], element_type: str) -> dict[st
         raise ValueError("Parameter 'na_color' must be color-like.")
 
     if (norm := param_dict.get("norm")) is not None:
-        if element_type in ["images", "labels"] and not isinstance(norm, Normalize):
+        if element_type in {"images", "labels"} and not isinstance(norm, Normalize):
             raise TypeError("Parameter 'norm' must be of type Normalize.")
         if element_type in ["shapes", "points"] and not isinstance(norm, bool | Normalize):
             raise TypeError("Parameter 'norm' must be a boolean or a mpl.Normalize.")
 
     if (scale := param_dict.get("scale")) is not None:
-        if element_type in ["images", "labels"] and not isinstance(scale, str):
+        if element_type in {"images", "labels"} and not isinstance(scale, str):
             raise TypeError("Parameter 'scale' must be a string if specified.")
         if element_type == "shapes":
             if not isinstance(scale, float | int):
@@ -1626,10 +1711,53 @@ def _type_check_params(param_dict: dict[str, Any], element_type: str) -> dict[st
         if size < 0:
             raise ValueError("Parameter 'size' must be a positive number.")
 
-    if param_dict.get("table_name") and not isinstance(param_dict["table_name"], str):
-        raise TypeError("Parameter 'table_name' must be a string .")
+    table_name = param_dict.get("table_name")
+    table_layer = param_dict.get("table_layer")
+    if table_name and not isinstance(param_dict["table_name"], str):
+        raise TypeError("Parameter 'table_name' must be a string.")
+
+    if table_layer and not isinstance(param_dict["table_layer"], str):
+        raise TypeError("Parameter 'table_layer' must be a string.")
+
+    def _ensure_table_and_layer_exist_in_sdata(
+        sdata: SpatialData, table_name: str | None, table_layer: str | None
+    ) -> bool:
+        """Ensure that table_name and table_layer are valid; throw error if not."""
+        if table_name:
+            if table_layer:
+                if table_layer in sdata.tables[table_name].layers:
+                    return True
+                raise ValueError(f"Layer '{table_layer}' not found in table '{table_name}'.")
+            return True  # using sdata.tables[table_name].X
+
+        if table_layer:
+            # user specified a layer but we have no tables => invalid
+            if len(sdata.tables) == 0:
+                raise ValueError("Trying to use 'table_layer' but no tables are present in the SpatialData object.")
+            if len(sdata.tables) == 1:
+                single_table_name = list(sdata.tables.keys())[0]
+                if table_layer in sdata.tables[single_table_name].layers:
+                    return True
+                raise ValueError(f"Layer '{table_layer}' not found in table '{single_table_name}'.")
+            # more than one tables, try to find which one has the given layer
+            found_table = False
+            for tname in sdata.tables:
+                if table_layer in sdata.tables[tname].layers:
+                    if found_table:
+                        raise ValueError(
+                            "Trying to guess 'table_name' based on 'table_layer', but found multiple matches."
+                        )
+                    found_table = True
+
+            if found_table:
+                return True
+
+            raise ValueError(f"Layer '{table_layer}' not found in any table.")
+
+        return True  # not using any table
+
+    assert _ensure_table_and_layer_exist_in_sdata(param_dict.get("sdata"), table_name, table_layer)
 
-    # like this because the following would assign True/False to 'method'
     if (method := param_dict.get("method")) not in ["matplotlib", "datashader", None]:
         raise ValueError("If specified, parameter 'method' must be either 'matplotlib' or 'datashader'.")
 
@@ -1668,6 +1796,7 @@ def _validate_label_render_params(
     outline_alpha: float | int,
     scale: str | None,
     table_name: str | None,
+    table_layer: str | None,
 ) -> dict[str, dict[str, Any]]:
     param_dict: dict[str, Any] = {
         "sdata": sdata,
@@ -1683,12 +1812,12 @@ def _validate_label_render_params(
         "norm": norm,
         "scale": scale,
         "table_name": table_name,
+        "table_layer": table_layer,
     }
     param_dict = _type_check_params(param_dict, "labels")
 
     element_params: dict[str, dict[str, Any]] = {}
     for el in param_dict["element"]:
-
         # ensure that the element exists in the SpatialData object
         _ = param_dict["sdata"][el]
 
@@ -1696,11 +1825,11 @@ def _validate_label_render_params(
         element_params[el]["na_color"] = param_dict["na_color"]
         element_params[el]["cmap"] = param_dict["cmap"]
         element_params[el]["norm"] = param_dict["norm"]
-        element_params[el]["color"] = param_dict["color"]
         element_params[el]["fill_alpha"] = param_dict["fill_alpha"]
         element_params[el]["scale"] = param_dict["scale"]
         element_params[el]["outline_alpha"] = param_dict["outline_alpha"]
         element_params[el]["contour_px"] = param_dict["contour_px"]
+        element_params[el]["table_layer"] = param_dict["table_layer"]
 
         element_params[el]["table_name"] = None
         element_params[el]["color"] = None
@@ -1727,6 +1856,7 @@ def _validate_points_render_params(
     norm: Normalize | None,
     size: float | int,
     table_name: str | None,
+    table_layer: str | None,
     ds_reduction: str | None,
 ) -> dict[str, dict[str, Any]]:
     param_dict: dict[str, Any] = {
@@ -1741,13 +1871,13 @@ def _validate_points_render_params(
         "norm": norm,
         "size": size,
         "table_name": table_name,
+        "table_layer": table_layer,
         "ds_reduction": ds_reduction,
     }
     param_dict = _type_check_params(param_dict, "points")
 
     element_params: dict[str, dict[str, Any]] = {}
     for el in param_dict["element"]:
-
         # ensure that the element exists in the SpatialData object
         _ = param_dict["sdata"][el]
 
@@ -1758,6 +1888,7 @@ def _validate_points_render_params(
         element_params[el]["color"] = param_dict["color"]
         element_params[el]["size"] = param_dict["size"]
         element_params[el]["alpha"] = param_dict["alpha"]
+        element_params[el]["table_layer"] = param_dict["table_layer"]
 
         element_params[el]["table_name"] = None
         element_params[el]["col_for_color"] = None
@@ -1790,6 +1921,7 @@ def _validate_shape_render_params(
     norm: Normalize | None,
     scale: float | int,
     table_name: str | None,
+    table_layer: str | None,
     method: str | None,
     ds_reduction: str | None,
 ) -> dict[str, dict[str, Any]]:
@@ -1808,6 +1940,7 @@ def _validate_shape_render_params(
         "norm": norm,
         "scale": scale,
         "table_name": table_name,
+        "table_layer": table_layer,
         "method": method,
         "ds_reduction": ds_reduction,
     }
@@ -1815,7 +1948,6 @@ def _validate_shape_render_params(
 
     element_params: dict[str, dict[str, Any]] = {}
     for el in param_dict["element"]:
-
         # ensure that the element exists in the SpatialData object
         _ = param_dict["sdata"][el]
 
@@ -1828,6 +1960,7 @@ def _validate_shape_render_params(
         element_params[el]["cmap"] = param_dict["cmap"]
         element_params[el]["norm"] = param_dict["norm"]
         element_params[el]["scale"] = param_dict["scale"]
+        element_params[el]["table_layer"] = param_dict["table_layer"]
 
         element_params[el]["color"] = param_dict["color"]
 
@@ -1849,9 +1982,12 @@ def _validate_shape_render_params(
 
 
 def _validate_col_for_column_table(
-    sdata: SpatialData, element_name: str, col_for_color: str | None, table_name: str | None, labels: bool = False
+    sdata: SpatialData,
+    element_name: str,
+    col_for_color: str | None,
+    table_name: str | None,
+    labels: bool = False,
 ) -> tuple[str | None, str | None]:
-
     if not labels and col_for_color in sdata[element_name].columns:
         table_name = None
     elif table_name is not None:
@@ -1871,7 +2007,11 @@ def _validate_col_for_column_table(
         elif len(tables) >= 1:
             table_name = next(iter(tables))
             if len(tables) > 1:
-                warnings.warn(f"Multiple tables contain color column, using {table_name}", UserWarning, stacklevel=2)
+                warnings.warn(
+                    f"Multiple tables contain column '{col_for_color}', using table '{table_name}'.",
+                    UserWarning,
+                    stacklevel=2,
+                )
     return col_for_color, table_name
 
 
@@ -1904,25 +2044,49 @@ def _validate_image_render_params(
         element_params[el] = {}
         spatial_element = param_dict["sdata"][el]
 
+        # robustly get channel names from image or multiscale image
         spatial_element_ch = (
-            spatial_element.c if isinstance(spatial_element, DataArray) else spatial_element["scale0"].c
+            spatial_element.c.values if isinstance(spatial_element, DataArray) else spatial_element["scale0"].c.values
         )
-        if (channel := param_dict["channel"]) is not None and (
-            (isinstance(channel[0], int) and max([abs(ch) for ch in channel]) <= len(spatial_element_ch))
-            or all(ch in spatial_element_ch for ch in channel)
-        ):
+        channel = param_dict["channel"]
+        if channel is not None:
+            # Normalize channel to always be a list of str or a list of int
+            if isinstance(channel, str):
+                channel = [channel]
+
+            if isinstance(channel, int):
+                channel = [channel]
+
+            # If channel is a list, ensure all elements are the same type
+            if not (isinstance(channel, list) and channel and all(isinstance(c, type(channel[0])) for c in channel)):
+                raise TypeError("Each item in 'channel' list must be of the same type, either string or integer.")
+
+            invalid = [c for c in channel if c not in spatial_element_ch]
+            if invalid:
+                raise ValueError(
+                    f"Invalid channel(s): {', '.join(str(c) for c in invalid)}. Valid choices are: {spatial_element_ch}"
+                )
             element_params[el]["channel"] = channel
         else:
             element_params[el]["channel"] = None
 
         element_params[el]["alpha"] = param_dict["alpha"]
 
-        if isinstance(palette := param_dict["palette"], list):
+        palette = param_dict["palette"]
+        assert isinstance(palette, list | type(None))  # if present, was converted to list, just to make sure
+
+        if isinstance(palette, list):
+            # case A: single palette for all channels
             if len(palette) == 1:
                 palette_length = len(channel) if channel is not None else len(spatial_element_ch)
                 palette = palette * palette_length
-            if (channel is not None and len(palette) != len(channel)) and len(palette) != len(spatial_element_ch):
-                palette = None
+            # case B: one palette per channel (either given or derived from channel length)
+            channels_to_use = spatial_element_ch if element_params[el]["channel"] is None else channel
+            if channels_to_use is not None and len(palette) != len(channels_to_use):
+                raise ValueError(
+                    f"Palette length ({len(palette)}) does not match channel length "
+                    f"({', '.join(str(c) for c in channels_to_use)})."
+                )
         element_params[el]["palette"] = palette
         element_params[el]["na_color"] = param_dict["na_color"]
 
@@ -1953,7 +2117,12 @@ def _get_wanted_render_elements(
     element_type: Literal["images", "labels", "points", "shapes"],
 ) -> tuple[list[str], list[str], bool]:
     wants_elements = True
-    if element_type in ["images", "labels", "points", "shapes"]:  # Prevents eval security risk
+    if element_type in [
+        "images",
+        "labels",
+        "points",
+        "shapes",
+    ]:  # Prevents eval security risk
         wanted_elements: list[str] = [params.element]
         wanted_elements_on_cs = [
             element for element in wanted_elements if cs in set(get_transformation(sdata[element], get_all=True).keys())
@@ -1971,18 +2140,37 @@ def _is_coercable_to_float(series: pd.Series) -> bool:
 
 
 def _ax_show_and_transform(
-    array: MaskedArray[tuple[int, ...], Any],
+    array: MaskedArray[tuple[int, ...], Any] | npt.NDArray[Any],
     trans_data: CompositeGenericTransform,
     ax: Axes,
     alpha: float | None = None,
     cmap: ListedColormap | LinearSegmentedColormap | None = None,
     zorder: int = 0,
+    extent: list[float] | None = None,
+    norm: Normalize | None = None,
 ) -> matplotlib.image.AxesImage:
+    # default extent in mpl:
+    image_extent = [-0.5, array.shape[1] - 0.5, array.shape[0] - 0.5, -0.5]
+    if extent is not None:
+        # make sure extent is [x_min, x_max, y_min, y_max]
+        if extent[3] < extent[2]:
+            extent[2], extent[3] = extent[3], extent[2]
+        if extent[0] < 0:
+            x_factor = array.shape[1] / (extent[1] - extent[0])
+            image_extent[0] = image_extent[0] + (extent[0] * x_factor)
+            image_extent[1] = image_extent[1] + (extent[0] * x_factor)
+        if extent[2] < 0:
+            y_factor = array.shape[0] / (extent[3] - extent[2])
+            image_extent[2] = image_extent[2] + (extent[2] * y_factor)
+            image_extent[3] = image_extent[3] + (extent[2] * y_factor)
+
     if not cmap and alpha is not None:
         im = ax.imshow(
             array,
             alpha=alpha,
             zorder=zorder,
+            extent=tuple(image_extent),
+            norm=norm,
         )
         im.set_transform(trans_data)
     else:
@@ -1990,6 +2178,8 @@ def _ax_show_and_transform(
             array,
             cmap=cmap,
             zorder=zorder,
+            extent=tuple(image_extent),
+            norm=norm,
         )
         im.set_transform(trans_data)
     return im
@@ -2018,7 +2208,10 @@ def set_zero_in_cmap_to_transparent(cmap: Colormap | str, steps: int | None = No
 
 
 def _get_extent_and_range_for_datashader_canvas(
-    spatial_element: SpatialElement, coordinate_system: str, ax: Axes, fig_params: FigParams
+    spatial_element: SpatialElement,
+    coordinate_system: str,
+    ax: Axes,
+    fig_params: FigParams,
 ) -> tuple[Any, Any, list[Any], list[Any], Any]:
     extent = get_extent(spatial_element, coordinate_system=coordinate_system)
     x_ext = [min(0, extent["x"][0]), extent["x"][1]]
@@ -2054,10 +2247,12 @@ def _get_extent_and_range_for_datashader_canvas(
 
 
 def _create_image_from_datashader_result(
-    ds_result: ds.transfer_functions.Image, factor: float, ax: Axes
-) -> tuple[MaskedArray[tuple[int, ...], Any], matplotlib.transforms.CompositeGenericTransform]:
+    ds_result: ds.transfer_functions.Image | np.ndarray[Any, np.dtype[np.uint8]],
+    factor: float,
+    ax: Axes,
+) -> tuple[MaskedArray[tuple[int, ...], Any], matplotlib.transforms.Transform]:
     # create SpatialImage from datashader output to get it back to original size
-    rgba_image_data = ds_result.to_numpy().base
+    rgba_image_data = ds_result.copy() if isinstance(ds_result, np.ndarray) else ds_result.to_numpy().base
     rgba_image_data = np.transpose(rgba_image_data, (2, 0, 1))
     rgba_image = Image2DModel.parse(
         rgba_image_data,
@@ -2138,7 +2333,7 @@ def _datashader_aggregate_with_function(
 
 
 def _datshader_get_how_kw_for_spread(
-    reduction: Literal["sum", "mean", "any", "count", "std", "var", "max", "min"] | None
+    reduction: Literal["sum", "mean", "any", "count", "std", "var", "max", "min"] | None,
 ) -> str:
     # Get the best input for the how argument of ds.tf.spread(), needed for numerical values
     reduction = reduction or "sum"
@@ -2163,27 +2358,151 @@ def _datshader_get_how_kw_for_spread(
     return reduction_to_how_map[reduction]
 
 
-def _robust_get_value(
-    sdata: sd.SpatialData,
-    origin: _ValueOrigin,
-    value_to_plot: str | None,
-    element_name: list[str] | str | None = None,
-    table_name: str | None = None,
-) -> pd.Series | None:
-    """Locate the value to plot in the spatial data object."""
-    model = get_model(sdata[element_name])
-    if model == PointsModel and table_name is not None:
-        return get_values_point_table(sdata=sdata, origin=origin, table_name=table_name)
-    vals = get_values(value_key=value_to_plot, sdata=sdata, element_name=element_name, table_name=table_name)
-    return vals[value_to_plot]
-
-
 def _prepare_transformation(
-    element: DataArray | GeoDataFrame | dask.dataframe.core.DataFrame, coordinate_system: str, ax: Axes | None = None
-) -> tuple[matplotlib.transforms.Affine2D, matplotlib.transforms.CompositeGenericTransform | None]:
+    element: DataArray | GeoDataFrame | dask.dataframe.core.DataFrame,
+    coordinate_system: str,
+    ax: Axes | None = None,
+) -> tuple[
+    matplotlib.transforms.Affine2D,
+    matplotlib.transforms.CompositeGenericTransform | None,
+]:
     trans = get_transformation(element, get_all=True)[coordinate_system]
     affine_trans = trans.to_affine_matrix(input_axes=("x", "y"), output_axes=("x", "y"))
     trans = mtransforms.Affine2D(matrix=affine_trans)
     trans_data = trans + ax.transData if ax is not None else None
 
     return trans, trans_data
+
+
+def _get_datashader_trans_matrix_of_single_element(
+    trans: Identity | Scale | Affine | MapAxis | Translation,
+) -> npt.NDArray[Any]:
+    flip_matrix = np.array([[1, 0, 0], [0, -1, 0], [0, 0, 1]])
+    tm: npt.NDArray[Any] = trans.to_affine_matrix(("x", "y"), ("x", "y"))
+
+    if isinstance(trans, Identity):
+        return np.array([[1, 0, 0], [0, 1, 0], [0, 0, 1]])
+    if isinstance(trans, (Scale | Affine)):
+        # idea: "flip the y-axis", apply transformation, flip back
+        flip_and_transform: npt.NDArray[Any] = flip_matrix @ tm @ flip_matrix
+        return flip_and_transform
+    if isinstance(trans, MapAxis):
+        # no flipping needed
+        return tm
+    # for a Translation, we need the transposed transformation matrix
+    tm_T = tm.T
+    assert isinstance(tm_T, np.ndarray)
+    return tm_T
+
+
+def _get_transformation_matrix_for_datashader(
+    trans: Scale | Identity | Affine | MapAxis | Translation | SDSequence,
+) -> npt.NDArray[Any]:
+    """Get the affine matrix needed to transform shapes for rendering with datashader."""
+    if isinstance(trans, SDSequence):
+        tm = np.array([[1, 0, 0], [0, 1, 0], [0, 0, 1]])
+        for x in trans.transformations:
+            tm = tm @ _get_datashader_trans_matrix_of_single_element(x)
+        return tm
+    return _get_datashader_trans_matrix_of_single_element(trans)
+
+
+def _datashader_map_aggregate_to_color(
+    agg: DataArray,
+    cmap: str | list[str] | ListedColormap,
+    color_key: None | list[str] = None,
+    min_alpha: float = 40,
+    span: None | list[float] = None,
+    clip: bool = True,
+) -> ds.tf.Image | np.ndarray[Any, np.dtype[np.uint8]]:
+    """ds.tf.shade() part, ensuring correct clipping behavior.
+
+    If necessary (norm.clip=False), split shading in 3 parts and in the end, stack results.
+    This ensures the correct clipping behavior, because else datashader would always automatically clip.
+    """
+    if not clip and isinstance(cmap, Colormap) and span is not None:
+        # in case we use datashader together with a Normalize object where clip=False
+        # why we need this is documented in https://github.com/scverse/spatialdata-plot/issues/372
+        agg_in = agg.where((agg >= span[0]) & (agg <= span[1]))
+        img_in = ds.tf.shade(
+            agg_in,
+            cmap=cmap,
+            span=(span[0], span[1]),
+            how="linear",
+            color_key=color_key,
+            min_alpha=min_alpha,
+        )
+
+        agg_under = agg.where(agg < span[0])
+        img_under = ds.tf.shade(
+            agg_under,
+            cmap=[to_hex(cmap.get_under())[:7]],
+            min_alpha=min_alpha,
+            color_key=color_key,
+        )
+
+        agg_over = agg.where(agg > span[1])
+        img_over = ds.tf.shade(
+            agg_over,
+            cmap=[to_hex(cmap.get_over())[:7]],
+            min_alpha=min_alpha,
+            color_key=color_key,
+        )
+
+        # stack the 3 arrays manually: go from under, through in to over and always overlay the values where alpha=0
+        stack = img_under.to_numpy().base
+        if stack is None:
+            stack = img_in.to_numpy().base
+        else:
+            stack[stack[:, :, 3] == 0] = img_in.to_numpy().base[stack[:, :, 3] == 0]
+        img_over = img_over.to_numpy().base
+        if img_over is not None:
+            stack[stack[:, :, 3] == 0] = img_over[stack[:, :, 3] == 0]
+        return stack
+
+    return ds.tf.shade(
+        agg,
+        cmap=cmap,
+        color_key=color_key,
+        min_alpha=min_alpha,
+        span=span,
+        how="linear",
+    )
+
+
+def _hex_no_alpha(hex: str) -> str:
+    """
+    Return a hex color string without an alpha component.
+
+    Parameters
+    ----------
+    hex : str
+        The input hex color string. Must be in one of the following formats:
+        - "#RRGGBB": a hex color without an alpha channel.
+        - "#RRGGBBAA": a hex color with an alpha channel that will be removed.
+
+    Returns
+    -------
+    str
+        The hex color string in "#RRGGBB" format.
+    """
+    if not isinstance(hex, str):
+        raise TypeError("Input must be a string")
+    if not hex.startswith("#"):
+        raise ValueError("Invalid hex color: must start with '#'")
+
+    hex_digits = hex[1:]
+    length = len(hex_digits)
+
+    if length == 6:
+        if not all(c in "0123456789abcdefABCDEF" for c in hex_digits):
+            raise ValueError("Invalid hex color: contains non-hex characters")
+        return hex  # Already in #RRGGBB format.
+
+    if length == 8:
+        if not all(c in "0123456789abcdefABCDEF" for c in hex_digits):
+            raise ValueError("Invalid hex color: contains non-hex characters")
+        # Return only the first 6 characters, stripping the alpha.
+        return "#" + hex_digits[:6]
+
+    raise ValueError("Invalid hex color length: must be either '#RRGGBB' or '#RRGGBBAA'")
